@@ -9,12 +9,24 @@ import webbrowser
 from keycloak import KeycloakOpenID
 from qiskit import qpy
 from qiskit import QuantumCircuit
+from qiskit.quantum_info import Operator, Pauli, PauliList, SparsePauliOp
+from qiskit.quantum_info.operators.linear_op import LinearOp
 from urllib.parse import urlencode
 
+
+def serialize_object(object):
+    buffer = io.BytesIO()
+    qpy.dump(object, buffer)
+    qpy_binary_data = buffer.getvalue()
+    base64_encoded_object = base64.b64encode(qpy_binary_data).decode("utf-8")
+    return base64_encoded_object
+
+
 class AuthenticationFailure(Exception):
     def __init__(self, message):
         self.message = message
 
+
 class Job:
     def __init__(self, job_id):
         self._job_id = job_id
@@ -22,6 +34,7 @@ class Job:
     def id(self):
         return self._job_id
 
+
 class WorkflowJob:
     def __init__(self, job_id):
         self._job_id = job_id
@@ -29,6 +42,7 @@ class WorkflowJob:
     def id(self):
         return self._job_id
 
+
 class InputData:
     def __init__(self, label=None, content=None):
         self.data = {}
@@ -36,36 +50,40 @@ class InputData:
             self.add_data(label, content)
 
     def add_data(self, label, content):
-        self.check_label(label, content)
-        if label == "pubs":
-            content = self.format_pubs(content)
-        if type(content) == dict or type(content) == list or type(content) == tuple:
-            self.data[label] = json.dumps(content, indent=4)
-        elif type(content) == str:
-            self.data[label] = content
-        else:
-            raise Exception("Input data content must be either a dictionary, or a list/tuple of JSON serializable objects, or a string.")
+        self.check_label(label, self.data)
+        try:
+            if label == "operator":
+                content = self.validate_and_serialize_operator(content)
+            elif label == "pub":
+                content = self.validate_and_serialize_pub(content)
+                if not "pubs" in self.data.keys():
+                    self.data["pubs"] = []
+                self.data["pubs"].append(content)
+            else:
+                self.data[label] = content
+        except (OverflowError, TypeError, ValueError):
+            raise Exception("Input data content must be JSON serializable.")
 
-    def check_label(self, label, content):
+    def check_label(self, label, data):
         if type(label) != str:
             raise Exception("Input data label must be string.")
-        if not label in ["pubs", "molecule-info"]:
-            raise Exception("Only 'pubs' or 'molecule-info' can be specified as a input data labels.")
-        if label == "pubs" and type(content) != list:
-            raise Exception("For a 'pubs' data label the content of data must be a list of PUBs. Each PUB must be a tuple containing a quantum circuit, optionally second a list of circuit parameters and optionally third a number of shots.")
-
-    def format_pubs(self, content):
-        pubs = []
-        for pub in content:
-            pubs.append(self.validate_and_serialize_pub(pub))
-        return pubs
-
-    def serialize_circuit(self, qc):
-        buffer = io.BytesIO()
-        qpy.dump(qc, buffer)
-        qpy_binary_data = buffer.getvalue()
-        base64_encoded_circuit = base64.b64encode(qpy_binary_data).decode("utf-8")
-        return base64_encoded_circuit
+        if label not in [
+            "ansatz-parameters",
+            "ising-model",
+            "lattice",
+            "molecule-info",
+            "operator",
+            "pub",
+            "qubo-model",
+            "vectors-data",
+        ]:
+            raise Exception(
+                f"Input data of type {label} is not supported. Please choose one of the following options: 'ising-model', 'lattice', 'molecule-info', 'operator', 'pub', 'qubo-model', 'data-vectors'."
+            )
+        if label != "pub" and label in data.keys():
+            raise Exception(
+                f"An input data item of type '{label}' has already been added to the job input data. Multiple data items of same category are allowed only for PUBs."
+            )
 
     def validate_and_serialize_pub(self, pub):
         shots = None
@@ -73,7 +91,9 @@ class InputData:
         if type(pub) == QuantumCircuit:
             quantum_circuit = pub
         elif type(pub) != tuple:
-            raise Exception("A pub can be either a quantum circuit or a tuple containing a quantum circuit, optionally second a list of circuit parameters and optionally third a number of shots.")
+            raise Exception(
+                "A pub can be either a quantum circuit or a tuple containing a quantum circuit, optionally second a list of circuit parameters and optionally third a number of shots."
+            )
         elif len(pub) == 3:
             quantum_circuit, paramaters, shots = pub
         elif len(pub) == 2:
@@ -81,19 +101,52 @@ class InputData:
         elif len(pub) == 1:
             quantum_circuit = pub[0]
         else:
-            raise Exception("A pub can be a tuple with at most 3 elements: a quantum circuit, a list of circuit paramaters and a number of shots.")
+            raise Exception(
+                "A pub can be a tuple with at most 3 elements: a quantum circuit, a list of circuit paramaters and a number of shots."
+            )
         if shots is not None and type(shots) != int:
-            raise Exception("The 'shots' setting in a PUB must be an integer and be positioned as the third element of a tuple specifying a PUB.")
+            raise Exception(
+                "The 'shots' setting in a PUB must be an integer and be positioned as the third element of a tuple specifying a PUB."
+            )
         if paramaters is not None and type(paramaters) != list:
-            raise Exception("The 'paramaters' in a PUB must be a list of numbers and be positioned as the second element of a tuple specifying a PUB.")
-        if quantum_circuit.num_parameters == 0 and (paramaters is not None and len(paramaters) != 0):
-            raise Exception("A circuit with zero parameters must have 'paramaters' argument 'None' or an empty list.")
-        elif paramaters is not None and quantum_circuit.num_parameters != len(paramaters):
-            raise Exception(f"The number of paramaters for a quantum circuit {quantum_circuit.num_parameters} is different from the length {len(paramaters)} of the list of aruguments.")
-        if paramaters is not None and not all(isinstance(item, (int, float)) for item in paramaters):
-            raise Exception("The 'paramaters' setting in a PUB must be a list of numbers.")
+            raise Exception(
+                "The 'paramaters' in a PUB must be a list of numbers and be positioned as the second element of a tuple specifying a PUB."
+            )
+        if quantum_circuit.num_parameters == 0 and (
+            paramaters is not None and len(paramaters) != 0
+        ):
+            raise Exception(
+                "A circuit with zero parameters must have 'paramaters' argument 'None' or an empty list."
+            )
+        elif paramaters is not None and quantum_circuit.num_parameters != len(
+            paramaters
+        ):
+            raise Exception(
+                f"The number of paramaters for a quantum circuit {quantum_circuit.num_parameters} is different from the length {len(paramaters)} of the list of aruguments."
+            )
+        if paramaters is not None and not all(
+            isinstance(item, (int, float)) for item in paramaters
+        ):
+            raise Exception(
+                "The 'paramaters' setting in a PUB must be a list of numbers."
+            )
+
+        return (serialize_object(quantum_circuit), paramaters, shots)
+
+
+def validate_and_serialize_operator(self, operator):
+    if (
+        not isinstance(operator, Operator)
+        and not isinstance(operator, Pauli)
+        and not isinstance(operator, PauliList)
+        and not isinstance(operator, SparsePauliOp)
+        and not isinstance(operator, LinearOp)
+    ):
+        raise Exception(
+            "The operator must be an instance of the Operator, Pauli, PauliList or SparsePauliOp class."
+        )
+    return self.serialize_object(operator)
 
-        return (self.serialize_circuit(quantum_circuit), paramaters, shots)
 
 class StrafulProvider:
 
@@ -158,6 +211,8 @@ In case the service has been recently started please wait 5 minutes for it to be
                 print("More details: ", ex.message)
         except Exception as ex:
             print("Failed to authenticate with the quantum provider.")
+            if "Connection refused" in str(ex):
+                print("The remote service does not respond. Please try again later.")
             if self._debug:
                 print("Unexpected exception: ", ex)
 
@@ -168,7 +223,7 @@ In case the service has been recently started please wait 5 minutes for it to be
             print("Please specify the backend name.")
             return
         if not circuit:
-            print("The circuit cannot be empty.")
+            print("The quantum circuit is missing.")
             return
         if shots is None:
             print("Please specify the number of shots.")
@@ -179,7 +234,7 @@ In case the service has been recently started please wait 5 minutes for it to be
         try:
             job_data = {
                 "BackendName": backend,
-                "CircuitData": circuit,
+                "Circuit": serialize_object(circuit),
                 "Shots": shots,
                 "Comments": comments,
             }
@@ -190,14 +245,20 @@ In case the service has been recently started please wait 5 minutes for it to be
                 return Job(result["id"])
             else:
                 print(
-                    f"Job submission has failed with http status code: {status_code}."
+                    f"Job submission has failed with http status code: {status_code}. \nRemote server response: '{result}'"
                 )
                 return Job(None)
         except Exception as ex:
             print(str(ex))
 
     def submit_workflow_job(
-        self, *, backend=None, shots=None, workflow_id=None, input_data=InputData(), comments=""
+        self,
+        *,
+        backend=None,
+        shots=None,
+        workflow_id=None,
+        input_data=InputData(),
+        comments="",
     ):
         if not self._verify_user_is_authenticated():
             return
@@ -224,7 +285,8 @@ In case the service has been recently started please wait 5 minutes for it to be
             input_data_items.append(str(shots))
             for input_data_label in input_data.data.keys():
                 input_data_labels.append(input_data_label)
-                input_data_items.append(input_data.data[input_data_label])
+                content = input_data.data[input_data_label]
+                input_data_items.append(json.dumps(content, indent=4))
             job_data = {
                 "BackendName": backend,
                 "WorkflowId": workflow_id,
@@ -239,7 +301,9 @@ In case the service has been recently started please wait 5 minutes for it to be
             if status_code == 201:
                 return WorkflowJob(result["id"])
             else:
-                print(f"Workflow job submission has failed: {result}.")
+                print(
+                    f"Workflow job submission has failed with http status code: {status_code}. \nRemote server response: '{result}'"
+                )
                 return WorkflowJob(None)
         except Exception as ex:
             print(str(ex))
@@ -322,7 +386,7 @@ In case the service has been recently started please wait 5 minutes for it to be
             or self._refresh_token_expiration_time is None
         ):
             print(
-                "You are not authorized to access this service. Please try to authenticate."
+                "You are not authorized to access this service. Please try to authenticate first."
             )
             return False
         if self.is_refresh_token_expired():
@@ -350,9 +414,9 @@ In case the service has been recently started please wait 5 minutes for it to be
         )
         try:
             json = response.json()
+            return (response.status_code, json)
         except:
-            json = "{}"
-        return (response.status_code, json)
+            return (response.status_code, response.text)
 
     def is_token_expired(self):
         if self._token_expiration_time is None:
@@ -404,7 +468,7 @@ In case the service has been recently started please wait 5 minutes for it to be
                 )
             else:
                 raise AuthenticationFailure(
-                    "Cannot store state to initiate authentication, provider does not respond."
+                    "Cannot initiate authentication, the authentication provider does not respond."
                 )
         self._state = state
 
